@@ -1,5 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2005, 2007 Free Software Foundation, Inc.
+   Copyright (C) 1999-2001, 2005, 2007, 2009-2012, 2014-2015 Free
+   Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,13 +13,11 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 /* This is a "fake" mta designed for testing purposes. It imitates
    sendmail sending and daemon modes. It does not actually send anything,
-   instead it just outputs the transcript of what would have been done.
+   instead it just outputs a transcript of what would have been done.
 
    Invocation:
    
@@ -28,13 +27,21 @@
    recipients.
 
    2. mta -bd [-p port]
-
+      mta -bD [-p port]
+      
    Operates as daemon. If port is given, mta will listen on that port.
    Otherwise, it will use the first free port in the range 1024-65535.
-   In this case, mta prints the port number on the stdout, prior to
-   starting operation. Notice, that in this mode mta does not disconnect
-   itself from the controlling terminal, it always stays on the foreground.
+   If the -bd option is given, mta switches to the background and prints
+   its PID and port number on two separate lines on the stdout.  It then
+   listens for incoming connections for no more than 60 seconds.
 
+   If the -bD option is given, the utility prints the port number and
+   listens for incoming connections not detaching itself from the controlling
+   terminal.
+
+   In any case, at most one connection is served, after which the utility
+   terminates.
+   
    Environment variables:
 
    MTA_DOMAIN   Sets the default domain name for email addresses. Default
@@ -56,23 +63,36 @@
 #endif
 #include <string.h>
 #include <pwd.h>
+#include <sysexits.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
-#include <mu_asprintf.h>
 #include <mailutils/argcv.h>
 #include <mailutils/mailutils.h>
 
 FILE *diag = NULL;       /* diagnostic output */
 int smtp_mode;           /* Operate in smtp mode */
 int port = 0;            /* Port number (for smtp mode) */
+int foreground = 0;      /* Stay in the foreground (smtp mode) */
 char *from_person = NULL; /* Set the name of the `from' person */
 int read_recipients = 0; /* Read the message for recipients */
 int dot = 1;             /* Message is terminated by a lone dot on a line */
 
 mu_address_t recipients = NULL;
-char *progname;
+
+/* FIXME: If finalize_option is set, mta should try to finalize
+   received messages the way sendmail does, i.e. to add To: or
+   Cc: headers, if they are missing, etc. The code to do so is
+   already included, but the modified message is not reproduced
+   on diagnostic output because mta_send reads it from the stream,
+   which does not reflect modifications to the header.
+
+   Ideally, the mu_message_get_streamref function should notice the
+   change in the header (and/or the body) and return a temporary
+   stream, which will read the modified values.  This is left as
+   as TODO for a later time.                  2010-09-29, Sergey */
+int finalize_option = 0; 
 
 int mta_stdin (int argc, char **argv);
 int mta_smtp (int argc, char **argv);
@@ -84,9 +104,7 @@ main (int argc, char **argv)
   int c, status;
   char *domain;
 
-  progname = strrchr (argv[0], '/');
-  if (!progname)
-    progname = argv[0];
+  mu_set_program_name (argv[0]);
   
   while ((c = getopt (argc, argv, "b:f:p:to:")) != EOF)
     {
@@ -99,6 +117,11 @@ main (int argc, char **argv)
 	      smtp_mode = 1;
 	      break;
 
+	    case 'D':
+	      smtp_mode = 1;
+	      foreground = 1;
+	      break;
+	      
 	    default:
 	      /*FIXME*/;
 	    }
@@ -129,7 +152,7 @@ main (int argc, char **argv)
 	  break;
 
 	default:
-	  exit (1);
+	  exit (EX_USAGE);
 	}
     }
 
@@ -142,13 +165,13 @@ main (int argc, char **argv)
 	  diag = fopen (name, mode);
 	  if (!diag)
 	    {
-	      mu_error ("%s: can't open diagnostic output: %s",
-			progname, name);
+	      mu_error ("can't open diagnostic output %s: %s",
+			name, mu_strerror (errno));
 	      return 1;
 	    }
 	}
       else
-	diag = stdout;
+	diag = stderr;
     }
   
   register_handlers ();
@@ -178,81 +201,83 @@ from_address ()
   return strdup (from_person);
 }
      
-void
-make_tmp (FILE *input, const char *from, char **tempfile)
+static mu_message_t
+make_tmp (mu_stream_t in)
 {
-  time_t t;
-  int fd = mu_tempfile (NULL, tempfile);
-  FILE *fp;
+  int rc;
+  mu_stream_t out;
   char *buf = NULL;
-  size_t n = 0;
-  int line;
+  size_t size = 0, n;
+  mu_message_t mesg;
   
-  if (fd == -1 || (fp = fdopen (fd, "w+")) == NULL)
+  rc = mu_temp_file_stream_create (&out, NULL, 0);
+  if (rc)
     {
-      mu_error ("%s: unable to open temporary file", progname);
-      exit (1);
+      mu_error (_("unable to open temporary file: %s"), mu_strerror (rc));
+      exit (EX_UNAVAILABLE);
     }
 
-  line = 0;
-  while (getline (&buf, &n, input) > 0)
+  rc = mu_stream_getline (in, &buf, &size, &n);
+  if (rc)
     {
-      int len = strlen (buf);
-      if (len >= 2 && buf[len - 2] == '\r')
+      mu_error (_("read error: %s"), mu_strerror (rc));
+      mu_stream_destroy (&out);
+      exit (EX_UNAVAILABLE);
+    }
+  if (n == 0)
+    {
+      mu_error (_("unexpected EOF on input"));
+      mu_stream_destroy (&out);
+      exit (EX_DATAERR);
+    } 
+
+  if (n >= 5 && memcmp (buf, "From ", 5))
+    {
+      time_t t;
+      struct tm *tm;
+      const char *from = from_address ();
+      if (!from)
 	{
-	  buf[len - 2] = '\n';
-	  buf[len - 1] = 0;
+	  mu_error ("can't determine sender address");
+	  exit (EX_NOUSER);
 	}
 	  
-      line++;
-      if (line == 1)
-	{
-	  if (memcmp (buf, "From ", 5))
-	    {
-	      char *from = from_address ();
-	      if (from)
-		{
-		  time (&t);
-		  fprintf (fp, "From %s %s", from, ctime (&t));
-		  free (from);
-		}
-	      else
-		{
-		  mu_error ("%s: can't determine sender address", progname);
-		  exit (1);
-		}
-	    }
-	}
-      else if (!memcmp (buf, "From ", 5))
-	fputc ('>', fp);
-
-      if (dot && buf[0] == '.' &&
-	       ((buf[1] == '\r' && buf[2] == '\n') || (buf[1] == '\n')))
-	break;
-      
-      if (fputs (buf, fp) == EOF)
-	{
-	  mu_error ("%s: temporary file write error", progname);
-	  fclose (fp);
-	  exit (1);
-	}
+      time (&t);
+      tm = gmtime (&t);
+      mu_stream_printf (out, "From %s ", from);
+      mu_c_streamftime (out, "%c%n", tm, NULL);
     }
-  
-  if (buf && strchr (buf, '\n') == NULL)
-    putc ('\n', fp);
 
-  putc ('\n', fp);
+  mu_stream_write (out, buf, n, NULL);
   free (buf);
+  
+  rc = mu_stream_copy (out, in, 0, NULL);
+  if (rc)
+    {
+      mu_error (_("copy error: %s"), mu_strerror (rc));
+      mu_stream_destroy (&out);
+      exit (EX_UNAVAILABLE);
+    }
 
-  fclose (fp);
+  rc = mu_stream_to_message (out, &mesg);
+  mu_stream_destroy (&out);
+  if (rc)
+    {
+      mu_error (_("error creating temporary message: %s"),
+		mu_strerror (rc));
+      exit (EX_UNAVAILABLE);
+    }
+
+  return mesg;
 }
 
 void
 register_handlers ()
 {
-  mu_registrar_record (mu_path_record);
+  mu_registrar_record (mu_mbox_record);
   mu_registrar_record (mu_sendmail_record);
   mu_registrar_record (mu_smtp_record);
+  mu_registrar_set_default_record (mu_mbox_record);
 }
 
 int
@@ -288,7 +313,7 @@ address_email_string (mu_address_t addr)
   value = malloc (length + 1);
   if (!value)
     {
-      mu_error ("%s: not enough memory", progname);
+      mu_error ("not enough memory");
       return NULL;
     }
   p = value;
@@ -311,9 +336,10 @@ mta_send (mu_message_t msg)
   size_t n;
   char buffer[512];    
   mu_stream_t stream = NULL;
-  size_t off = 0, line;
+  size_t line;
   char *value;
-
+  int newline = 0;
+  
   value = from_address ();
   if (value)
     {
@@ -325,26 +351,35 @@ mta_send (mu_message_t msg)
   fprintf (diag, "ENVELOPE TO: %s\n", value);
   free (value);
 
-  mu_message_get_stream (msg, &stream);
+  mu_message_get_streamref (msg, &stream);
   line = 0;
-  fprintf (diag, "%4lu: ", (unsigned long) line);
-  while (mu_stream_read (stream, buffer, sizeof buffer - 1, off, &n) == 0
+  newline = 1;
+  while (mu_stream_read (stream, buffer, sizeof buffer - 1, &n) == 0
 	 && n != 0)
     {
       size_t i;
 
       for (i = 0; i < n; i++)
 	{
+	  if (newline)
+	    {
+	      fprintf (diag, "%4lu:", (unsigned long) line);
+	      if (buffer[i] != '\n')
+		fputc (' ', diag);
+	      newline = 0;
+	    }
 	  fputc (buffer[i], diag);
 	  if (buffer[i] == '\n')
 	    {
 	      line++;
-	      fprintf (diag, "%4lu: ", (unsigned long) line);
+	      newline = 1;
 	    }
 	}
-      off += n;
     }
-  fprintf (diag, "\nEND OF MESSAGE\n");
+  mu_stream_destroy (&stream);
+  if (!newline)
+    fprintf (diag, "\n(no newline at EOF)\n");
+  fprintf (diag, "END OF MESSAGE\n");
   fflush (diag);
   return 0;
 }
@@ -360,14 +395,14 @@ message_finalize (mu_message_t msg, int warn)
   
   mu_message_get_header (msg, &header);
 
-  if (warn && from_person)
+  if (finalize_option && warn && from_person)
     {
       struct passwd *pwd = getpwuid (getuid ());
       char *warn = malloc (strlen (pwd->pw_name) + 1 +
 			   sizeof (SENDER_WARNING));
       if (warn == NULL)
 	{
-	  mu_error ("%s: not enough memory", progname);
+	  mu_error ("not enough memory");
 	  return 1;
 	}
       sprintf (warn, "%s %s", pwd->pw_name, SENDER_WARNING);
@@ -383,7 +418,7 @@ message_finalize (mu_message_t msg, int warn)
 	{
 	  if (add_recipient (value))
 	    {
-	      mu_error ("%s: bad address %s", progname, value);
+	      mu_error ("bad address %s", value);
 	      return 1;
 	    }
 	  free (value);
@@ -393,17 +428,18 @@ message_finalize (mu_message_t msg, int warn)
 	{
 	  if (add_recipient (value))
 	    {
-	      mu_error ("%s: bad address %s", progname, value);
+	      mu_error ("bad address %s", value);
 	      return 1;
 	    }
 	  free (value);
 	}  
 	  
-      if (mu_header_aget_value (header, MU_HEADER_BCC, &value) == 0)
+      if (finalize_option &&
+	  mu_header_aget_value (header, MU_HEADER_BCC, &value) == 0)
 	{
 	  if (add_recipient (value))
 	    {
-	      mu_error ("%s: bad address %s", progname, value);
+	      mu_error ("bad address %s", value);
 	      return 1;
 	    }
 	  free (value);
@@ -411,28 +447,19 @@ message_finalize (mu_message_t msg, int warn)
 	}  
     }
 
-  if (!have_to)
+  if (finalize_option && !have_to)
     {
-      size_t n;
+      char const *sptr;
       int c;
       
-      c = mu_address_to_string (recipients, NULL, 0, &n);
+      c = mu_address_sget_printable (recipients, &sptr);
       if (c)
 	{
-	  mu_error ("%s: mu_address_to_string failure: %s",
-		    progname, mu_strerror (c));
+	  mu_error ("mu_address_sget_printable failure: %s",
+		    mu_strerror (c));
 	  return 1;
 	}
-      value = malloc (n + 1);
-      if (!value)
-	{
-	  mu_error ("%s: not enough memory", progname);
-	  return 1;
-	}
-
-      mu_address_to_string (recipients, value, n + 1, &n);
-      mu_header_set_value (header, MU_HEADER_TO, value, 1);
-      free (value);
+      mu_header_set_value (header, MU_HEADER_TO, sptr, 1);
     }
   return 0;
 }
@@ -440,66 +467,53 @@ message_finalize (mu_message_t msg, int warn)
 int
 mta_stdin (int argc, char **argv)
 {
-  int c;
-  char *tempfile;
-  mu_mailbox_t mbox;
+  int c, rc;
+  mu_stream_t in;
   mu_message_t msg = NULL;
   
   for (c = 0; c < argc; c++)
     {
       if (add_recipient (argv[c]))
 	{
-	  mu_error ("%s: bad address %s", progname, argv[c]);
+	  mu_error ("bad address %s", argv[c]);
 	  return 1;
 	}
     }
 
-  make_tmp (stdin, from_person, &tempfile);
-  if ((c = mu_mailbox_create_default (&mbox, tempfile)) != 0)
+  rc = mu_stdio_stream_create (&in, MU_STDIN_FD, MU_STREAM_READ);
+  if (rc)
     {
-      mu_error ("%s: can't create mailbox %s: %s",
-		progname, tempfile, mu_strerror (c));
-      unlink (tempfile);
-      return 1;
-    }
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_stdio_stream_create",
+		       "MU_STDIN_FD", rc);
+      exit (EX_UNAVAILABLE);
+    } 
 
-  if ((c = mu_mailbox_open (mbox, MU_STREAM_RDWR)) != 0)
-    {
-      mu_error ("%s: can't open mailbox %s: %s",
-		progname, tempfile, mu_strerror (c));
-      unlink (tempfile);
-      return 1;
-    }
-
-  mu_mailbox_get_message (mbox, 1, &msg);
+  msg = make_tmp (in);
+  mu_stream_destroy (&in);
   if (message_finalize (msg, 1))
     return 1;
-
+  
   if (!recipients)
     {
-      mu_error ("%s: Recipient names must be specified", progname);
+      mu_error ("recipient names must be specified");
       return 1;
     }
 
   mta_send (msg);
-  
-  unlink (tempfile);
-  free (tempfile);
+  mu_message_destroy (&msg, mu_message_get_owner (msg));
   return 0;
 }
 
-FILE *in, *out;
-
 void
-smtp_reply (int code, char *fmt, ...)
+smtp_reply (mu_stream_t str, int code, char *fmt, ...)
 {
   va_list ap;
 
+  mu_stream_printf (str, "%d ", code);
   va_start (ap, fmt);
-  fprintf (out, "%d ", code);
-  vfprintf (out, fmt, ap);
+  mu_stream_vprintf (str, fmt, ap);
   va_end (ap);
-  fprintf (out, "\r\n");
+  mu_stream_printf (str, "\n");
 }
 
 #define STATE_INIT   0
@@ -537,7 +551,7 @@ smtp_kw (const char *name)
   int i;
 
   for (i = 0; kw[i].name != NULL; i++)
-    if (strcasecmp (name, kw[i].name) == 0)
+    if (mu_c_strcasecmp (name, kw[i].name) == 0)
       return kw[i].code;
   return -1;
 }
@@ -546,51 +560,52 @@ static char *
 check_prefix (char *str, const char *prefix)
 {
   int pflen = strlen (prefix);
-  if (strlen (str) > pflen && strncasecmp (str, prefix, pflen) == 0)
+  if (strlen (str) > pflen && mu_c_strncasecmp (str, prefix, pflen) == 0)
     return str + pflen;
   else
     return NULL;
 }
 
 void
-smtp (int fd)
+smtp (mu_stream_t str)
 {
-  int state, c;
+  int state;
   char *buf = NULL;
   size_t size = 0;
-  mu_mailbox_t mbox;
-  mu_message_t msg;
-  char *tempfile;
   char *rcpt_addr;
-  
-  in = fdopen (fd, "r");
-  out = fdopen (fd, "w");
-  SETVBUF (in, NULL, _IOLBF, 0);
-  SETVBUF (out, NULL, _IOLBF, 0);
+  int wsflags = MU_WRDSF_DEFFLAGS;
+  struct mu_wordsplit ws;
 
-  smtp_reply (220, "Ready");
+  smtp_reply (str, 220, "Ready");
   for (state = STATE_INIT; state != STATE_QUIT; )
     {
-      int argc;
-      char **argv;
-      int kw, len;
-      
-      if (getline (&buf, &size, in) == -1)
-	exit (1);
-      len = strlen (buf);
-      while (len > 0 && (buf[len-1] == '\n' || buf[len-1] == '\r'))
-	len --;
-      buf[len] = 0;
-      
-      if (mu_argcv_get (buf, "", NULL, &argc, &argv))
-	exit (1);
+      int rc;
+      int kw;
+      size_t len;
 
-      kw = smtp_kw (argv[0]);
+      rc = mu_stream_getline (str, &buf, &size, &len);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_stream_getline", NULL, rc);
+	  exit (EX_PROTOCOL);
+	}
+      
+      if (len == 0)
+	exit (EX_PROTOCOL);
+
+      if (mu_wordsplit (buf, &ws, wsflags))
+	{
+	  mu_error ("cannot split line `%s': %s", buf,
+		    mu_wordsplit_strerror (&ws));
+	  exit (EX_UNAVAILABLE);
+	}
+      wsflags |= MU_WRDSF_REUSE;
+
+      kw = smtp_kw (ws.ws_wordv[0]);
       if (kw == KW_QUIT)
 	{
-	  smtp_reply (221, "Done");
+	  smtp_reply (str, 221, "Done");
 	  state = STATE_QUIT;
-	  mu_argcv_free (argc, argv);
 	  continue;
 	}
       
@@ -601,17 +616,18 @@ smtp (int fd)
 	    {
 	    case KW_EHLO:
 	    case KW_HELO:
-	      if (argc == 2)
+	      if (ws.ws_wordc == 2)
 		{
-		  smtp_reply (250, "pleased to meet you");
+		  smtp_reply (str, 250, "pleased to meet you");
 		  state = STATE_EHLO;
 		}
 	      else
-		smtp_reply (501, "%s requires domain address", argv[0]);
+		smtp_reply (str, 501, "%s requires domain address",
+			    ws.ws_wordv[0]);
 	      break;
 
 	    default:
-	      smtp_reply (503, "Polite people say HELO first");
+	      smtp_reply (str, 503, "Polite people say HELO first");
 	      break;
 	    }
 	  break;
@@ -620,25 +636,26 @@ smtp (int fd)
 	  switch (kw)
 	    {
 	    case KW_MAIL:
-	      if (argc == 2)
-		from_person = check_prefix (argv[1], "from:");
-	      else if (argc == 3 && strcasecmp (argv[1], "from:") == 0)
-		from_person = argv[2];
+	      if (ws.ws_wordc == 2)
+		from_person = check_prefix (ws.ws_wordv[1], "from:");
+	      else if (ws.ws_wordc == 3 &&
+		       mu_c_strcasecmp (ws.ws_wordv[1], "from:") == 0)
+		from_person = ws.ws_wordv[2];
 	      else
 		from_person = NULL;
 
 	      if (from_person)
 		{
 		  from_person = strdup (from_person);
-		  smtp_reply (250, "Sender OK");
+		  smtp_reply (str, 250, "Sender OK");
 		  state = STATE_MAIL;
 		}
 	      else
-		smtp_reply (501, "Syntax error");
+		smtp_reply (str, 501, "Syntax error");
 	      break;
 
 	    default:
-	      smtp_reply (503, "Need MAIL command");
+	      smtp_reply (str, 503, "Need MAIL command");
 	    }
 	  break;
 	  
@@ -646,29 +663,30 @@ smtp (int fd)
 	  switch (kw)
 	    {
 	    case KW_RCPT:
-	      if (argc == 2)
-		rcpt_addr = check_prefix (argv[1], "to:");
-	      else if (argc == 3 && strcasecmp (argv[1], "to:") == 0)
-		rcpt_addr = argv[2];
+	      if (ws.ws_wordc == 2)
+		rcpt_addr = check_prefix (ws.ws_wordv[1], "to:");
+	      else if (ws.ws_wordc == 3 &&
+		       mu_c_strcasecmp (ws.ws_wordv[1], "to:") == 0)
+		rcpt_addr = ws.ws_wordv[2];
 	      else
 		rcpt_addr = NULL;
 	      
 	      if (rcpt_addr)
 		{
 		  if (add_recipient (rcpt_addr))
-		    smtp_reply (451, "Recipient not accepted");
+		    smtp_reply (str, 451, "Recipient not accepted");
 		  else
 		    {
-		      smtp_reply (250, "Recipient OK");
+		      smtp_reply (str, 250, "Recipient OK");
 		      state = STATE_RCPT;
 		    }
 		}
 	      else
-		smtp_reply (501, "Syntax error");
+		smtp_reply (str, 501, "Syntax error");
 	      break;
 	      
 	    default:
-	      smtp_reply (503, "Need RCPT command");
+	      smtp_reply (str, 503, "Need RCPT command");
 	    }
 	  break;
 	  
@@ -676,87 +694,86 @@ smtp (int fd)
 	  switch (kw)
 	    {
 	    case KW_RCPT:
-	      if (argc == 2)
-		rcpt_addr = check_prefix (argv[1], "to:");
-	      else if (argc == 3 && strcasecmp (argv[1], "to:") == 0)
-		rcpt_addr = argv[2];
+	      if (ws.ws_wordc == 2)
+		rcpt_addr = check_prefix (ws.ws_wordv[1], "to:");
+	      else if (ws.ws_wordc == 3 &&
+		       mu_c_strcasecmp (ws.ws_wordv[1], "to:") == 0)
+		rcpt_addr = ws.ws_wordv[2];
 	      else
 		rcpt_addr = NULL;
 	      
 	      if (rcpt_addr)
 		{
 		  if (add_recipient (rcpt_addr))
-		    smtp_reply (451, "Recipient not accepted");
+		    smtp_reply (str, 451, "Recipient not accepted");
 		  else
 		    {
-		      smtp_reply (250, "Recipient OK");
+		      smtp_reply (str, 250, "Recipient OK");
 		      state = STATE_RCPT;
 		    }
 		}
 	      else
-		smtp_reply (501, "Syntax error");
+		smtp_reply (str, 501, "Syntax error");
 	      break;
 
 	    case KW_DATA:
-	      smtp_reply (354,
-			  "Enter mail, end with \".\" on a line by itself");
-	      make_tmp (in, from_person, &tempfile);
-	      if ((c = mu_mailbox_create_default (&mbox, tempfile)) != 0)
-		{
-		  mu_error ("%s: can't create mailbox %s: %s",
-			    progname,
-			    tempfile, mu_strerror (c));
-		  unlink (tempfile);
-		  exit (1);
-		}
-
-	      if ((c = mu_mailbox_open (mbox, MU_STREAM_RDWR)) != 0)
-		{
-		  mu_error ("%s: can't open mailbox %s: %s",
-			    progname, 
-			    tempfile, mu_strerror (c));
-		  unlink (tempfile);
-		  exit (1);
-		}
-
-	      mu_mailbox_get_message (mbox, 1, &msg);
-	      if (message_finalize (msg, 0) == 0)
-		mta_send (msg);
-	      else
-		smtp_reply (501, "can't send message"); /*FIXME: code?*/
-	      unlink (tempfile);
-
-	      mu_address_destroy (&recipients);
-	      from_person = NULL;
+	      {
+		mu_stream_t flt;
+		mu_message_t msg;
+		int rc;
+		
+		rc = mu_filter_create (&flt, str, "DOT", MU_FILTER_DECODE,
+				       MU_STREAM_READ);
+		if (rc)
+		  {
+		    mu_diag_funcall (MU_DIAG_ERROR, "mu_filter_create",
+				     "CRLFDOT", rc);
+		    exit (EX_UNAVAILABLE);
+		  }
+		smtp_reply (str, 354,
+			    "Enter mail, end with \".\" on a line by itself");
+		
+		msg = make_tmp (flt);
+		mu_stream_destroy (&flt);
+		if (message_finalize (msg, 0) == 0)
+		  mta_send (msg);
+		else
+		  smtp_reply (str, 501, "can't send message"); /*FIXME: code?*/
+		mu_message_destroy (&msg, mu_message_get_owner (msg));
+		mu_address_destroy (&recipients);
+		from_person = NULL;
 	      
-	      smtp_reply (250, "Message accepted for delivery");
-	      state = STATE_EHLO;
+		smtp_reply (str, 250, "Message accepted for delivery");
+		state = STATE_EHLO;
+	      }
 	      break;
 
 	    default:
-	      smtp_reply (503, "Invalid command");
+	      smtp_reply (str, 503, "Invalid command");
 	      break;
 	    }
 	  break;
 
 	}
-      mu_argcv_free (argc, argv);
     }
-    
-  close (fd);
 }
 
+#ifndef INADDR_LOOPBACK
+# define INADDR_LOOPBACK 0x7f000001
+#endif
+
 int
 mta_smtp (int argc, char **argv)
 {
   int on = 1;
   struct sockaddr_in address;
   int fd;
+  pid_t pid;
   
   fd = socket (PF_INET, SOCK_STREAM, 0);
   if (fd < 0)
     {
-      mu_error ("%s: socket: %s", progname, strerror (errno));
+      mu_error ("socket: %s", mu_strerror (errno));
       return 1;
     }
 
@@ -764,7 +781,7 @@ mta_smtp (int argc, char **argv)
 
   memset (&address, 0, sizeof (address));
   address.sin_family = AF_INET;
-  address.sin_addr.s_addr = INADDR_ANY;
+  address.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
 
   if (port)
     {
@@ -772,7 +789,7 @@ mta_smtp (int argc, char **argv)
       if (bind (fd, (struct sockaddr *) &address, sizeof (address)) < 0)
 	{
 	  close (fd);
-	  mu_error ("%s: bind: %s", progname, strerror(errno));
+	  mu_error ("bind: %s", mu_strerror (errno));
 	  return 1;
 	}
     }
@@ -785,8 +802,7 @@ mta_smtp (int argc, char **argv)
 	{
 	  if (++port >= 65535)
 	    {
-	      mu_error ("%s: can't bind socket: all ports in use?",
-			progname);
+	      mu_error ("can't bind socket: all ports in use?");
 	      return 1;
 	    }
 	  address.sin_port = htons (port);
@@ -796,14 +812,39 @@ mta_smtp (int argc, char **argv)
     }
 
   listen (fd, 5);
-  printf ("%d\n", port);
+  if (!foreground)
+    {
+      pid = fork ();
+      if (pid == -1)
+	{
+	  mu_error ("fork: %s", mu_strerror (errno));
+	  return 1;
+	}
+      if (pid)
+	{
+	  /* Master process. */
+	  printf ("%lu\n", (unsigned long) pid);
+	  printf ("%d\n", port);
+	  return 0;
+	}
+    }
+  else
+    printf ("%d\n", port);
+  
+  /* Child */
   fclose (stdout);
+  if (!foreground)
+    alarm (60);
+  
   while (1)
     {
       fd_set rfds;
       struct sockaddr_in his_addr;
-      int sfd, len, status;
-
+      int sfd, status;
+      socklen_t len;
+      int rc;
+      mu_stream_t str, istream, ostream, flt;
+      
       FD_ZERO (&rfds);
       FD_SET (fd, &rfds);
       
@@ -812,21 +853,67 @@ mta_smtp (int argc, char **argv)
 	{
 	  if (errno == EINTR)
 	    continue;
-	  mu_error ("%s: select: %s", progname, strerror (errno));
+	  mu_error ("select: %s", mu_strerror (errno));
 	  return 1;
 	}
 
       len = sizeof (his_addr);
       if ((sfd = accept (fd, (struct sockaddr *)&his_addr, &len)) < 0)
 	{
-	  mu_error ("%s: accept: %s", progname, strerror (errno));
+	  mu_error ("accept: %s", mu_strerror (errno));
 	  return 1;
 	}
 
-      smtp (sfd);
+      rc = mu_stdio_stream_create (&istream, sfd, MU_STREAM_READ);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_stdio_stream_create",
+			   "MU_STREAM_READ", rc);
+	  return 1;
+	}
+      mu_stream_set_buffer (istream, mu_buffer_line, 0);
+      rc = mu_filter_create (&flt, istream, "CRLF", MU_FILTER_DECODE,
+			     MU_STREAM_READ);
+      mu_stream_unref (istream);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_filter_create",
+			   "CRLF,MU_FILTER_DECODE", rc);
+	  return 1;
+	}
+      mu_stream_set_buffer (flt, mu_buffer_line, 0);
+      istream = flt;
+
+      rc = mu_stdio_stream_create (&ostream, sfd, MU_STREAM_WRITE);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_stdio_stream_create",
+			   "MU_STREAM_WRITE", rc);
+	  return 1;
+	}
+      mu_stream_set_buffer (ostream, mu_buffer_line, 0);
+      rc = mu_filter_create (&flt, ostream, "CRLF", MU_FILTER_ENCODE,
+			     MU_STREAM_WRITE);
+      mu_stream_unref (ostream);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_filter_create",
+			   "CRLF,MU_FILTER_ENCODE", rc);
+	  return 1;
+	}
+      mu_stream_set_buffer (flt, mu_buffer_line, 0);
+      ostream = flt;
+      rc = mu_iostream_create (&str, istream, ostream);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_iostream_create", NULL, rc);
+	  return 1;
+	}
+      mu_stream_set_buffer (str, mu_buffer_line, 0);
+      smtp (str);
+      mu_stream_destroy (&str);
       break;
     }
   
   return 0;
 }
-
