@@ -1,7 +1,6 @@
-/*
-   GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2004, 2006, 2007, 2009, 2010 Free Software Foundation,
-   Inc.
+/* GNU Mailutils -- a suite of utilities for electronic mail
+   Copyright (C) 2004, 2006-2007, 2009-2012, 2014-2015 Free Software
+   Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -14,10 +13,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA
-*/
+   Public License along with this library.  If not, see 
+   <http://www.gnu.org/licenses/>. */
 
 #ifndef _MUCPP_POP3_H
 #define _MUCPP_POP3_H
@@ -51,7 +48,7 @@ class Pop3
 
   void apop (const char* name, const char* digest);
   void stls ();
-  Iterator& capa ();
+  Iterator& capa (bool reread);
   void dele (unsigned int msgno);
   size_t list (unsigned int msgno);
   Iterator& list_all ();
@@ -60,16 +57,15 @@ class Pop3
   void quit ();
   Stream& retr (unsigned int msgno);
   void rset ();
-  void stat (unsigned int* count, size_t* octets);
+  void stat (unsigned int* count, mu_off_t* octets);
   Stream& top (unsigned int msgno, unsigned int lines);
   std::string uidl  (unsigned int msgno);
   Iterator& uidl_all ();
   void user (const char* user);
 
-  size_t readline (char* buf, size_t buflen);
-  size_t response (char* buf, size_t buflen);
+  void getline ();
+  void response (size_t* nread);
   void sendline (const char* line);
-  void send ();
 };
 
 }
