@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2003, 2004, 2007, 2009, 2011 Free
-   Software Foundation, Inc.
+   Copyright (C) 1999-2001, 2003-2004, 2007, 2009, 2011-2012, 2014-2015
+   Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -107,8 +107,10 @@ int _mu_pop3_init (mu_pop3_t pop3);
         case EAGAIN:                            \
         case EINPROGRESS:                       \
         case EINTR:                             \
+	  return status;                        \
         case MU_ERR_REPLY:                      \
         case MU_ERR_BADREPLY:                   \
+	  pop3->state = MU_POP3_NO_STATE;	\
 	  return status;                        \
         default:                                \
           pop3->state = MU_POP3_ERROR;		\
